#
# canvtree.py
# $Id: canvtree.py,v 1.2 1996/09/09 16:42:00 kzeil Exp $
#
# Koert Zeilstra - Network Solutions
# 1996.09.05
#

import string
import Tkinter

# --------------------------------------------------
# TreeItem
# --------------------------------------------------
class TreeItem:
  def __init__ (self, label=None, label_tags=None,
	image=None, image_tags=None, subtree=None):
    self._label = label
    self._label_tags = label_tags
    self._image = image
    self._image_tags = image_tags
    self._subtree = subtree
    self._expandable = 0
    self._expand_callback = None

  def setImage (self, image):
    self._image = image
  def setSubtree(self, subtree):
    self._subtree = subtree

  def getLabel(self):
    return self._label
  def getLabelTags(self):
    return self._label_tags
  def getImage(self):
    return self._image
  def getImageTags(self):
    return self._image_tags
  def getSubtree(self):
    return self._subtree
  def getExpandable(self):
    return self._expandable
  def getExpandCallback(self):
    return self._expand_callback

  def makeExpandable(self, callback):
    self._expandable = 1
    self._expand_callback = callback

# --------------------------------------------------
# Tree
# --------------------------------------------------
class Tree:
  def __init__ (self):
    self._items = []
    self._colors = {"read": "yellow", "normal": "black",
		"select": "blue"}
    self._type_font = "-*-lucida-medium-r-*-*-*-120-*-*-*-*-*-*"
    self._line_height = 12
    self._x_indent = 15
    self._x_margin = 30
    self._y_margin = 5

  def addField (self, tags, image, label):
    item = TreeItem(tags, image, label)
    self._items.append(item)

  def addSubTree (self, tags, image, label, tree):
    item = TreeItem(tags, image, label, tree)
    self._items.append(item)

  def addItem (self, item):
    self._items.append(item)

  def getHeight (self):
    height = (len(self._items)+1) * (self._line_height + self._y_margin)
    for item in self._items:
      subtree = item.getSubtree()
      if subtree != None:
        height = height + subtree.getHeight()
    return height

  def display (self, canvas, x, y):
    for item in self._items:
      item_image = item.getImage()
      if item_image:
        item_id = canvas.create_image(x-15, y, image=item_image,
		anchor="nw", tags=item.getImageTags())
      item_id = canvas.create_text(x, y, text=item.getLabel(), anchor="nw",
		font=self._type_font, fill=self._colors["normal"],
		tags=item.getLabelTags())
      y = y + self._line_height + self._y_margin
      if item.getSubtree():
        x_dummy, y = item.getSubtree().display(canvas, x+15, y)
    return x, y
